#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#include "bud.h"
#include "common.h"
#include <pthread.h>

#if !defined(__x86_64) && !defined(__amd64) && !defined(__i386)
static pthread_mutex_t count_mutex;
#else
#pragma message "not using pthread_mutex_t"
#endif

void init_log(void)
{
#if !defined(__x86_64) && !defined(__amd64) && !defined(__i386)
    pthread_mutex_init(&count_mutex, NULL);
#endif
}

void uninit_log(void)
{
#if !defined(__x86_64) && !defined(__amd64) && !defined(__i386)
    pthread_mutex_destroy(&count_mutex);
#endif
}

/*helper functions*/

static inline int interlocked_xchg_add( int *dest, int incr )
{
    int ret;

#if defined(__x86_64) || defined(__amd64) || defined(__i386)
    {
        __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (incr) : "memory" );
    }
#else
    {
        pthread_mutex_lock(&count_mutex);
        ret = *dest;
        *dest += incr;
        pthread_mutex_unlock(&count_mutex);
    }
#endif

    return ret;
}

static char* get_temp_buffer(unsigned int size)
{
    #define MAX_LIST 64
    static char* list[MAX_LIST];
    static int pos = 0;
    int idx = interlocked_xchg_add(&pos,1) % MAX_LIST;
    char* buf = (char*)realloc(list[idx], size);
    if (buf)
    {
        list[idx] = buf;
    }

    return buf;
}

#define FUNC_ZERO(ast, sign) do {\
    *walk++ = '(';\
    walk = ast2str(ast->left, walk);\
    *walk++ = sign;\
    walk = ast2str(ast->right, walk);\
    *walk++ = ')';\
}while(0)

#define FUNC_ONE(ast, prefix) do {\
    walk += sprintf(walk, prefix);\
    *walk++ = '(';\
    walk = ast2str(ast->left, walk);\
    *walk++ = ')';\
}while(0)

#define FUNC_TWO(ast, prefix) do {\
    walk += sprintf(walk, prefix);\
    *walk++ = '(';\
    walk = ast2str(ast->left, walk);\
    *walk++ = ',';\
    walk = ast2str(ast->right, walk);\
    *walk++ = ')';\
}while(0)


static char* ast2str(struct tagAST* ast, char* strbuf)
{
    char* walk = strbuf;

    switch (ast->nodetype)
    {
    case '+': FUNC_ZERO(ast, '+');break;
    case '-': FUNC_ZERO(ast, '-');break;
    case '*': FUNC_ZERO(ast, '*');break;
    case '/': FUNC_ZERO(ast, '/');break;

    case 'N': FUNC_ONE(ast, "-");break;
    case 'R': FUNC_ONE(ast, "max_repeat");break;

    case '%': FUNC_TWO(ast, "mod");break;
    case 'm': FUNC_TWO(ast, "min");break;
    case 'M': FUNC_TWO(ast, "max");break;

    case 'K': walk += sprintf(walk, "%d", ((struct tagNumNode*)ast)->number); break;
    case 'V': walk += sprintf(walk, "%s", ((struct tagVarNode*)ast)->name); break;

    default:
        dbg_log(LOG_LEVEL_WARN, "bad node %c\n", ast->nodetype);
        break;
    }

    *walk = 0;

    return walk;
}

static char* cmd2hexstr(char* cmd)
{
    int i = 0;
    char* buf = get_temp_buffer(1024);
    unsigned char* walk = (unsigned char*)cmd+1;
    unsigned char count = *(unsigned char*)cmd;

    for (unsigned int j = 0; j<count; ++j)
    {
        sprintf(buf+i,"%02X ", *walk);
        i += 3;
        ++walk;
    }
    buf[i] = 0;

    return buf;
}

static char* arg2str(struct tagArgList* arg)
{
    char* expr = get_temp_buffer(1024*4);
    char* buf = expr + 512;
    char* walk = buf;

    struct tagCmdArg* head = arg->table;

    for (char i = 0; i<arg->count; ++i)
    {
        if ('s' == head->type)
        {
            walk += sprintf(walk, "%s", cmd2hexstr((char*)(head->ast)));
        }
        else
        {
            ast2str(head->ast, expr);

            /*TODO: call free_ast(head->ast); to free memory*/

            if (head->max > head->min)
                walk += sprintf(walk, "%%%c[%d,%d]{%s}", head->type, head->min, head->max, expr);
            else
                walk += sprintf(walk, "%%%c{%s}", head->type, expr);
        }

        ++head;
    }

    return buf;
}

static char* numlist2str(struct tagNumList* head)
{
    char* buf = get_temp_buffer(4096);
    char* walk = buf;

    *walk = 0;

    for (unsigned int i = 0; i<head->count; ++i)
    {
        walk += sprintf(walk, "%d,",head->table[i]);
    }

    return buf;
}

static char* strlist2str(struct tagStrList* head)
{
    char* buf = get_temp_buffer(4096);
    char* walk = buf;

    *walk = 0;

    for (unsigned int i = 0; i<head->count; ++i)
    {
        walk += sprintf(walk, "%s,", head->table[i]);
    }

    return buf;
}


static const char* get_level(unsigned int level)
{
    char* level_buffer = get_temp_buffer(256);

    for (unsigned int i = 0; i<level; ++i)
    {
        level_buffer[2*i+0] = ' ';
        level_buffer[2*i+1] = ' ';
    }

    level_buffer[2*level] = 0;

    return level_buffer;
}

static char* attribute2str(struct tagAttribute* attr)
{
    char* buf = get_temp_buffer(4096);

    buf[0] = 0;
    switch (attr->type)
    {
    case GPD_NUMBER:
       sprintf(buf, "%d", attr->num);
       break;
    case GPD_STRING:
       return attr->str;
    case GPD_PAIR:
       if (!strcmp(attr->name, "DPI"))
            sprintf(buf, "%dx%ddpi", attr->pair.first, attr->pair.second);
       else
            sprintf(buf, "PAIR(%d,%d)", attr->pair.first, attr->pair.second);
       break;
    case GPD_RECT:
       sprintf(buf, "RECT(%d,%d,%d,%d)",
               attr->rect.left, attr->rect.top, attr->rect.right, attr->rect.bottom);
       break;
    case GPD_LIST:
       if (attr->ilist.type == 'K')
         sprintf(buf, "LIST(%s)", numlist2str(&attr->ilist));
       else if (attr->ilist.type == 'c')
         sprintf(buf, "LIST(%s)", strlist2str(&attr->slist));
       break;
    case GPD_EXPR:
       sprintf(buf, "%s",  arg2str(&attr->arg));
       break;
    case GPD_MACRO:
       break;
    default:
       assert(0);
       break;
    }

    return buf;
}

int print_attributes(struct tagAttribute* attr, int count)
{
    for (int i = 0; i<count; ++i)
    {
        fprintf(stderr, "%s:%s\n", attr[i].name, attribute2str(&attr[i]));
    }
}

static int print_entry_list(struct tagEntryList** head, int count, int level)
{
    for (int i =0; i<count; ++i)
    {
        struct tagEntryList* it = head[i];
        switch (it->type)
        {
        case 'a':
        {
            char* str = NULL;
            str = attribute2str((struct tagAttribute*)it);

            fprintf(stderr, "%s%s:%s\n", get_level(level),
                   ((struct tagAttribute*)it)->name, str);
        }
        break;

        case 'e':
        {
            struct tagEntry* entry = (struct tagEntry*)it;

            switch (entry->type)
            {
            case ENTRY_FEATURE:
                fprintf(stderr, "%sFeature:%s\n", get_level(level), entry->name);
                break;
            case ENTRY_OPTION:
                fprintf(stderr, "%sOption:%s\n", get_level(level), entry->name);
                break;
            case ENTRY_COMMAND:
                fprintf(stderr, "%sCommand:%s\n", get_level(level), entry->name);
                break;
            case ENTRY_SWITCH:
                fprintf(stderr, "%sswitch:%s\n", get_level(level), entry->name);
                break;
            case ENTRY_CASE:
                fprintf(stderr, "%scase:%s\n", get_level(level), entry->name);
                break;
            case ENTRY_FONTCAR:
                fprintf(stderr, "%sFontCartridge:%s\n", get_level(level), entry->name);
                break;
            default:
               fprintf(stderr, "%s Not support entry, type:%d, ptr:%p\n",
                       get_level(level), entry->type, entry);
               break;
            }

            if (entry->sub)
            {
                if (ENTRY_SWITCH == entry->type || ENTRY_CASE == entry->type)
                    print_entry_list (entry->sub, entry->count, level);
                else
                    print_entry_list (entry->sub, entry->count, level + 1);
            }
        }
        break;

        default:
        dbg_log(LOG_LEVEL_WARN, "%s Not support node, type:%c, ptr:%p\n",
                get_level(level), it->type, it);
        assert(0);
        break;
        }
    }
}

int print_entries(struct tagEntry* entry, int count)
{
    for (int i = 0; i<count; ++i)
    {
        struct tagEntryList* ptr = (struct tagEntryList*)&entry[i];
        print_entry_list(&ptr, 1, 0);
    }
}

/*end*/
