#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include "cups/cups.h"
#include "cups/raster.h"
#include "cups/i18n.h"
#include "jpeglib.h"
#include "raster_converter.h"

#define SAFE_BYTES_PER_LINE(w, bpp)  (((w)*(bpp)+31)&(~31))/8

static void save_bmp(unsigned char* buffer, int bytes, const char* prefix,
        int id, int width, int height, int bpp)
{
#ifdef _DEBUG
    char name[128];
    sprintf(name, "/tmp/%s-%d-%dx%d-%dbpp.bin", prefix, id, width, height, bpp);
    FILE* pf = fopen(name, "wb");
    if (pf)
    {
        fwrite(buffer, 1, bytes, pf);
        fclose(pf);
    }
#endif
}

/*keep vertical size not changed*/
static void clip_image_horizontal_8bpp(unsigned char* buffer, int width, int height,
        int left, int right, unsigned char* out )
{
    unsigned char* walk = out;
    unsigned char* line = buffer;

    for (int y = 0; y<height; ++y)
    {
        for (int x = left; x<right; ++x)
        {
            *walk++ = line[x];
        }

        line += width;
    }
}

/*keep vertical size not changed*/
static void find_smallest_print_area_8bpp(unsigned char* buffer,
        int width, int height, int* left, int* right)
{
    int w = width;
    int h = height;
    int l = w-1;
    int r = 1;
    unsigned char* line = buffer;

    for (int y = 0; y<h; ++y)
    {
        for (int x = 0; x<w; ++x)
        {
            if (line[x])
            {
                if (x < l) l = x;
                else if (x > r) r = x;
            }
        }

        line += w;
    }

    *left = l;
    *right = r;
}

static void rotate_90_8bpp(unsigned char* buffer, int width, int height, unsigned char* out)
{
    unsigned char* walk = NULL;
    unsigned char* end = buffer + width*height;

    for (int x = 0; x<width; ++x)
    {
        walk = end + x;
        for (int y = height; y>0; --y)
        {
            walk -= width;
            *out++ = *walk;
        }
    }
}

static int do_black8bpp_to_black1bpp( cups_page_header2_t* header, unsigned char* buffer,
        unsigned char* clip_buffer, unsigned char* rotate_buffer,
        int lines_of_block, int lines_of_parts, RasterBlock* out)
{
    /*clip white area in horizontal and vertical*/
    int left = 0, right = 0;
    int parts_of_block = lines_of_block/lines_of_parts;
    assert (parts_of_block < 5);

    find_smallest_print_area_8bpp(buffer,
            header->cupsWidth, lines_of_block, &left, &right);
    if (left > right)
    {
        /*remove whole white area*/
        return -1;
    }

    /*FIXME: Always keep the number of lines is multiple times of lines_of_parts */
    out->x = left;
    out->cols = right - left + 1;
    out->rows = lines_of_block;
    out->data_bytes = lines_of_block*SAFE_BYTES_PER_LINE(out->cols, 1);
    out->data = (unsigned char*)malloc(out->data_bytes);

    if (rotate_buffer)
    {
        unsigned int shift = 0;
        unsigned char* src = NULL;
        unsigned char* dest = out->data;
        unsigned char* line = rotate_buffer;
        int width = out->cols;
        int height = out->rows;
        int offset = (parts_of_block-1)*lines_of_parts;

        if (out->cols < header->cupsWidth)
        {
            clip_image_horizontal_8bpp(buffer,
                header->cupsWidth, height,
                left, left+width, clip_buffer);

            rotate_90_8bpp(clip_buffer, width, height, rotate_buffer);
        }
        else
        {
            rotate_90_8bpp(buffer, width, height, rotate_buffer);
        }

        for (int y = 0; y<width; ++y)
        {
            src = line + offset;
            for (int j = parts_of_block-1; j>=0; --j)
            {
                for (int x = 0; x<lines_of_parts; ++x)
                {
                    if (src[x]>0)
                        dest[0] |= (1<<shift);
                    else
                        dest[0] &= ~(1<<shift);

                    ++shift;
                    if (shift > 7)
                    {
                        shift = 0;
                        ++dest;
                    }
                }
                src -= lines_of_parts;
            }

            line += lines_of_block;
        }

        out->data_bytes = (unsigned int)(dest - out->data);
        if (shift > 0)
        {
            ++out->data_bytes;
        }

        save_bmp(out->data, out->data_bytes, "xx", out->y, width, height, 1);
    }
    else
    {
#ifdef _DEBUG
        fprintf(stderr, "FIXME: horizontal raster not supported!\n");
#endif
    }

    return 0;
}

static int convert_black8bpp_to_black1bpp(cups_raster_t* ras, cups_page_header2_t* header,
        bool horizontal, int lines_of_block, int parts_of_block,
        RasterBlock** raster_block, int* raster_block_count)
{
    int last_y = 0;
    int lines_of_parts = lines_of_block/parts_of_block;
    unsigned int buffer_len = header->cupsBytesPerLine*lines_of_block;
    unsigned char* buffer = (unsigned char*)malloc(buffer_len*3);
    unsigned char* clip_buffer = buffer + buffer_len;
    unsigned char* rotate_buffer = horizontal ? NULL: (clip_buffer + buffer_len);
    int block_count = (header->cupsHeight + lines_of_block - 1)/lines_of_block;
    RasterBlock* out = (RasterBlock*)malloc(block_count*sizeof(RasterBlock));

    assert(0 == (lines_of_block%parts_of_block));
    *raster_block = out;
    *raster_block_count = block_count;

    for (int i = 0; i<block_count; ++i)
    {
        out[i].x = 0;
        out[i].y = last_y;
        out[i].color_fmt = Black_1_BPP;
        out[i].data_bytes = 0;
        out[i].data = NULL;

        if (cupsRasterReadPixels(ras, buffer, buffer_len) < 1)
        {
            /*last block, read line by line*/
            unsigned char* walk = buffer;
            int remain = header->cupsHeight - last_y;
            for (int j = 0; i<remain; ++j)
            {
                if (cupsRasterReadPixels(ras,
                                         walk,
                                         header->cupsBytesPerLine) < 1)
                {
                    _cupsLangPrintFilter(stderr, "ERROR",
                             _("failed to read raster pixel data:%d,%d,%d"),
                             i, j, remain);

                    assert(0);
                }
                walk += header->cupsBytesPerLine;
            }
            memset(walk, 0, (lines_of_block - remain)*header->cupsBytesPerLine);
        }

        do_black8bpp_to_black1bpp(header, buffer, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, out+i);

        last_y += lines_of_block;
    }

    free(buffer);

    return 0;
}

static void interlacing_8bpp_block(unsigned char* buffer, int width,
    int height, unsigned char* even_buffer, unsigned char* odd_buffer)
{
    unsigned char* even = buffer;
    unsigned char* odd = buffer + width;

    for (int y = 0; y<height; ++y)
    {
        for (int x = 0; x<width; ++x)
        {
            even_buffer[x] = even[x];
            odd_buffer[x] = odd[x];
        }

        even += width*2;
        odd += width*2;

        even_buffer += width;
        odd_buffer += width;
    }
}

static int convert_black8bpp_to_black1bpp_interlacing(cups_raster_t* ras, cups_page_header2_t* header,
        bool horizontal, int lines_of_block, int parts_of_block,
        RasterBlock** raster_block, int* raster_block_count)
{
    int last_y = 0;
    int lines_of_parts = lines_of_block/parts_of_block;
    unsigned int buffer_len = header->cupsBytesPerLine*lines_of_block;
    unsigned char* buffer = (unsigned char*)malloc(buffer_len*6);
    unsigned char* clip_buffer = buffer + buffer_len*2;
    unsigned char* even_buffer = clip_buffer + buffer_len;
    unsigned char* odd_buffer = even_buffer + buffer_len;
    unsigned char* rotate_buffer = horizontal ? NULL: (odd_buffer + buffer_len);
    int block_count = (header->cupsHeight + lines_of_block - 1)/lines_of_block;
    RasterBlock* out = (RasterBlock*)malloc(block_count*sizeof(RasterBlock));
    RasterBlock* even = out;
    RasterBlock* odd = out+1;

    assert(0 == (lines_of_block%parts_of_block));
    *raster_block = out;
    *raster_block_count = block_count;

    for (int i = 0; i<block_count; i+= 2)
    {
        /*even parts*/
        even->x = 0;
        even->y = last_y;
        even->color_fmt = Black_1_BPP;
        even->data_bytes = 0;
        even->data = NULL;

        /*odd parts*/
        odd->x = 0;
        odd->y = last_y+2; /*FIXME: hard code move down 2 rows is right?*/
        odd->color_fmt = Black_1_BPP;
        odd->data_bytes = 0;
        odd->data = NULL;

        if (cupsRasterReadPixels(ras, buffer, buffer_len*2) < 1)
        {
            /*last block, read line by line*/

            unsigned char* walk = buffer;
            int remain = header->cupsHeight - last_y;
            for (int j = 0; i<remain; ++j)
            {
                if (cupsRasterReadPixels(ras,
                                         walk,
                                         header->cupsBytesPerLine) < 1)
                {
                    _cupsLangPrintFilter(stderr, "ERROR",
                             _("failed to read raster pixel data:%d,%d,%d"),
                             i, j, remain);

                    assert(0);
                }
                walk += header->cupsBytesPerLine;
            }
            memset(walk, 0, (lines_of_block*2 - remain)*header->cupsBytesPerLine);
        }

        interlacing_8bpp_block(buffer, header->cupsBytesPerLine,
                lines_of_block, even_buffer, odd_buffer);

        do_black8bpp_to_black1bpp(header, even_buffer, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, even);

        do_black8bpp_to_black1bpp(header, odd_buffer, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, odd);

        even += 2;
        odd += 2;

        last_y += lines_of_block*2;
    }

    free(buffer);

    return 0;
}

static void black1bpp_to_black8bpp(unsigned char* buffer,
        int width, int height, unsigned char* row)
{
    int shift = 0;
    unsigned char* walk = buffer;
    for (int y = 0; y<height; ++y)
    {
        shift = 7;
        for (int x = 0; x<width; ++x)
        {
            if (walk[0]&(1<<shift))
            {
                row[x] = 0x1;
            }
            else
            {
                row[x] = 0x0;
            }

            --shift;
            if (shift<0)
            {
                shift = 7;
                ++walk;
            }
        }

        if (shift<7)
        {
            ++walk;
        }

        row += width;
    }
}

static int convert_black1bpp_to_black1bpp(cups_raster_t* ras, cups_page_header2_t* header,
        bool horizontal, int lines_of_block, int parts_of_block,
        RasterBlock** raster_block, int* raster_block_count)
{
    int last_y = 0;
    int lines_of_parts = lines_of_block/parts_of_block;
    unsigned int buffer_len = header->cupsBytesPerLine*lines_of_block;
    unsigned int buffer_len8bpp = (header->cupsWidth*8)*lines_of_block;
    unsigned char* buffer = (unsigned char*)malloc(buffer_len8bpp*3+buffer_len);
    unsigned char* buffer8bpp = buffer + buffer_len;
    unsigned char* clip_buffer = buffer8bpp + buffer_len8bpp;
    unsigned char* rotate_buffer = horizontal ? NULL: (clip_buffer + buffer_len8bpp);
    int block_count = (header->cupsHeight + lines_of_block - 1)/lines_of_block;
    RasterBlock* out = (RasterBlock*)malloc(block_count*sizeof(RasterBlock));

    assert(0 == (lines_of_block%parts_of_block));
    *raster_block = out;
    *raster_block_count = block_count;

    for (int i = 0; i<block_count; ++i)
    {
        out[i].x = 0;
        out[i].y = last_y;
        out[i].color_fmt = Black_1_BPP;
        out[i].data_bytes = 0;
        out[i].data = NULL;

        if (cupsRasterReadPixels(ras, buffer, buffer_len) < 1)
        {
            /*last block, read line by line*/
            unsigned char* walk = buffer;
            int remain = header->cupsHeight - last_y;
            for (int j = 0; i<remain; ++j)
            {
                if (cupsRasterReadPixels(ras,
                                         walk,
                                         header->cupsBytesPerLine) < 1)
                {
                    _cupsLangPrintFilter(stderr, "ERROR",
                             _("failed to read raster pixel data:%d,%d,%d"),
                             i, j, remain);

                    assert(0);
                }
                walk += header->cupsBytesPerLine;
            }
            memset(walk, 0, (lines_of_block - remain)*header->cupsBytesPerLine);
        }

        black1bpp_to_black8bpp(buffer, header->cupsWidth, lines_of_block, buffer8bpp);

        do_black8bpp_to_black1bpp(header, buffer8bpp, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, out+i);

        last_y += lines_of_block;
    }

    free(buffer);

    return 0;
}

static int convert_black1bpp_to_black1bpp_interlacing(cups_raster_t* ras, cups_page_header2_t* header,
        bool horizontal, int lines_of_block, int parts_of_block,
        RasterBlock** raster_block, int* raster_block_count)
{
    int last_y = 0;
    int lines_of_parts = lines_of_block/parts_of_block;
    unsigned int buffer_len = header->cupsBytesPerLine*lines_of_block;
    unsigned int buffer_len8bpp = (header->cupsWidth*8)*lines_of_block;
    unsigned char* buffer = (unsigned char*)malloc(buffer_len8bpp*6 + buffer_len*2);
    unsigned char* buffer8bpp = buffer + buffer_len*2;
    unsigned char* clip_buffer = buffer8bpp + buffer_len8bpp*2;
    unsigned char* even_buffer = clip_buffer + buffer_len8bpp;
    unsigned char* odd_buffer = even_buffer + buffer_len8bpp;
    unsigned char* rotate_buffer = horizontal ? NULL: (odd_buffer + buffer_len8bpp);
    int block_count = (header->cupsHeight + lines_of_block - 1)/lines_of_block;
    RasterBlock* out = (RasterBlock*)malloc(block_count*sizeof(RasterBlock));
    RasterBlock* even = out;
    RasterBlock* odd = out+1;

    assert(0 == (lines_of_block%parts_of_block));
    *raster_block = out;
    *raster_block_count = block_count;

    for (int i = 0; i<block_count; i+= 2)
    {
        /*even parts*/
        even->x = 0;
        even->y = last_y;
        even->color_fmt = Black_1_BPP;
        even->data_bytes = 0;
        even->data = NULL;

        /*odd parts*/
        odd->x = 0;
        odd->y = last_y+2; /*FIXME: hard code move down 2 rows is right?*/
        odd->color_fmt = Black_1_BPP;
        odd->data_bytes = 0;
        odd->data = NULL;

        if (cupsRasterReadPixels(ras, buffer, buffer_len*2) < 1)
        {
            /*last block, read line by line*/

            unsigned char* walk = buffer;
            int remain = header->cupsHeight - last_y;
            for (int j = 0; i<remain; ++j)
            {
                if (cupsRasterReadPixels(ras,
                                         walk,
                                         header->cupsBytesPerLine) < 1)
                {
                    _cupsLangPrintFilter(stderr, "ERROR",
                             _("failed to read raster pixel data:%d,%d,%d"),
                             i, j, remain);

                    assert(0);
                }
                walk += header->cupsBytesPerLine;
            }
            memset(walk, 0, (lines_of_block*2 - remain)*header->cupsBytesPerLine);
        }

        black1bpp_to_black8bpp(buffer, header->cupsWidth, lines_of_block*2, buffer8bpp);

        interlacing_8bpp_block(buffer8bpp, header->cupsBytesPerLine,
                lines_of_block, even_buffer, odd_buffer);

        do_black8bpp_to_black1bpp(header, even_buffer, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, even);

        do_black8bpp_to_black1bpp(header, odd_buffer, clip_buffer,
                rotate_buffer, lines_of_block, lines_of_parts, odd);

        even += 2;
        odd += 2;

        last_y += lines_of_block*2;
    }

    free(buffer);

    return 0;
}

int convert_raster_block(cups_raster_t* ras, cups_page_header2_t* header,
        ColorFmt color_fmt, bool horizontal, bool interlacing,
        int lines_of_block, int parts_of_block,
        RasterBlock** raster_block, int* raster_block_count)
{
    if (0)
    {
        fprintf(stderr, "DEBUG: Duplex = %d\n", header->Duplex);
        fprintf(stderr, "DEBUG: HWResolution = [ %d %d ]\n",
                header->HWResolution[0], header->HWResolution[1]);
        fprintf(stderr, "DEBUG: ImagingBoundingBox = [ %d %d %d %d ]\n",
                header->ImagingBoundingBox[0],
                header->ImagingBoundingBox[1],
                header->ImagingBoundingBox[2],
                header->ImagingBoundingBox[3]);
        fprintf(stderr, "DEBUG: Margins = [ %d %d ]\n", header->Margins[0], header->Margins[1]);
        fprintf(stderr, "DEBUG: ManualFeed = %d\n", header->ManualFeed);
        fprintf(stderr, "DEBUG: MediaPosition = %d\n", header->MediaPosition);
        fprintf(stderr, "DEBUG: NumCopies = %d\n", header->NumCopies);
        fprintf(stderr, "DEBUG: Orientation = %d\n", header->Orientation);
        fprintf(stderr, "DEBUG: PageSize = [ %d %d ]\n", header->PageSize[0],
                header->PageSize[1]);
        fprintf(stderr, "DEBUG: cupsWidth = %d\n", header->cupsWidth);
        fprintf(stderr, "DEBUG: cupsHeight = %d\n", header->cupsHeight);
        fprintf(stderr, "DEBUG: cupsMediaType = %d\n", header->cupsMediaType);
        fprintf(stderr, "DEBUG: cupsBitsPerColor = %d\n", header->cupsBitsPerColor);
        fprintf(stderr, "DEBUG: cupsBitsPerPixel = %d\n", header->cupsBitsPerPixel);
        fprintf(stderr, "DEBUG: cupsBytesPerLine = %d\n", header->cupsBytesPerLine);
        fprintf(stderr, "DEBUG: cupsColorOrder = %d\n", header->cupsColorOrder);
        fprintf(stderr, "DEBUG: cupsColorSpace = %d\n", header->cupsColorSpace);
        fprintf(stderr, "DEBUG: cupsCompression = %d\n", header->cupsCompression);

        unsigned int buffer_len = header->cupsBytesPerLine*1;
        unsigned char* line_buffer = (unsigned char*)malloc(buffer_len*header->cupsHeight);
        unsigned char* walk = line_buffer;

        for (unsigned int ret = 0; ;)
        {
            ret = cupsRasterReadPixels(ras, walk, buffer_len);
            if (ret < 1)
            {
                fprintf(stderr, "cupsRasterReadPixels failed %d\n", ret);
                break;
            }
            walk += buffer_len;
        }
        save_bmp(line_buffer, buffer_len*header->cupsHeight,
                "ras", 0, header->cupsWidth, header->cupsHeight, 1);

        return 0;
    }

    /*TODO: need process header->Orientation
      	0 = Do not rotate page
        1 = Rotate page counter-clockwise
        2 = Turn page upside down
        3 = Rotate page clockwise
    */

    switch (color_fmt)
    {
    case Black_1_BPP:
        {
            switch (header->cupsColorSpace)
            {
            case CUPS_CSPACE_K:
                switch (header->cupsBitsPerPixel)
                {
                case 1:
                if (!interlacing)
                    return convert_black1bpp_to_black1bpp(ras, header, horizontal,
                        lines_of_block, parts_of_block, raster_block, raster_block_count);
                else
                    return convert_black1bpp_to_black1bpp_interlacing(ras, header, horizontal,
                        lines_of_block, parts_of_block, raster_block, raster_block_count);
                    break;
                case 8:
                if (!interlacing)
                    return convert_black8bpp_to_black1bpp(ras, header, horizontal,
                        lines_of_block, parts_of_block, raster_block, raster_block_count);
                else
                    return convert_black8bpp_to_black1bpp_interlacing(ras, header, horizontal,
                        lines_of_block, parts_of_block, raster_block, raster_block_count);
                break;

                default:
                    _cupsLangPrintFilter(stderr, "ERROR",
                                 _("Unsupported color bits:%d, should change cupsBitsPerColor in ppd file!"),
                                 header->cupsBitsPerPixel);
                    break;
                }
                break;

            default:
                _cupsLangPrintFilter(stderr, "ERROR",
                             _("Unsupported raster color space: %d"),
                             header->cupsColorSpace);
                break;
            }
        }
        break;

    default:
        _cupsLangPrintFilter(stderr, "ERROR",
                 _("Unsupported target color space: %d"),
                 color_fmt);
        break;
    }

    return -1;
}

static int next_differ(unsigned char* seed, unsigned char* row, unsigned int offset,
        unsigned int len, unsigned int* diff_cnt)
{
    int ret = -1;
    unsigned int diff = 0;
    for (int i = 0; i < len; ++i) {
        if (seed[offset + i] == row[offset + i]) {
            if (!diff)
                continue;
            break;
        }
        if (!diff)
            ret = i;
        if (++diff == 8)
            break;
    }

    if (diff_cnt) *diff_cnt = diff;

    return ret;
}

typedef void (*convert_line_func)(unsigned char* raster_data,
        unsigned int len, unsigned char* data);

static void convert_line_k_8to1(unsigned char* raster_data, unsigned int len,
        unsigned char* data)
{
    unsigned char* orig = data;
    const int threshold = 0;
    memset(data, 0, (len + 7) / 8);
    for (unsigned int i = 0, j = 0; i < len; ++i)
    {
        if (raster_data[i] > threshold)
        {
            *data |= (unsigned char)0x80 >> (i & 0x07);
        }
        if ((i & 0x07) == 0x07)
            ++data;
    }
}

static void convert_line_k_1to1(unsigned char* raster_data, unsigned int len,
        unsigned char* data)
{
    memcpy(data, raster_data, len);
}

int convert_raster_block_pclxl(cups_raster_t* ras,
        cups_page_header2_t* header, RasterBlock** raster_block,
        int* raster_block_count)
{
    convert_line_func convert_line = NULL;
    int bpp = 1;
    ColorFmt color_fmt;

    if (header->cupsColorSpace == CUPS_CSPACE_K)
    {
        color_fmt = Black_1_BPP;
        if (header->cupsBitsPerColor == 1)
        {
            convert_line = &convert_line_k_1to1;
        }
        else if (header->cupsBitsPerColor == 8)
        {
            convert_line = &convert_line_k_8to1;
        }
        else
        {
            _cupsLangPrintFilter(stderr, "ERROR",
                    _("Unsupported color bits:%d, should change cupsBitsPerColor in ppd file!"),
                    header->cupsBitsPerPixel);
            return -1;
        }
    }
    else if (header->cupsColorSpace == CUPS_CSPACE_RGB)
    {
        color_fmt = RGB_24_BPP;
        bpp = 24;
        if (header->cupsBitsPerColor == 8)
        {
        }
        else
        {
            _cupsLangPrintFilter(stderr, "ERROR",
                    _("Unsupported color bits:%d, should change cupsBitsPerColor in ppd file!"),
                    header->cupsBitsPerPixel);
            return -1;
        }
    }
    else
    {
        _cupsLangPrintFilter(stderr, "ERROR",
                _("Unsupported raster color space: %d"),
                header->cupsColorSpace);
        return -1;
    }

    RasterBlock *out = (RasterBlock*)malloc(sizeof(RasterBlock));
    *raster_block = out;
    *raster_block_count = 1;
    out->x = 0;
    out->y = 0;
    out->rows = header->cupsHeight;
    out->cols = header->cupsWidth;
    out->color_fmt = color_fmt;
    out->data = NULL;
    out->data_bytes = 0;

    unsigned int line_len = header->cupsBytesPerLine;
    unsigned char* line_buffer = (unsigned char*)malloc(line_len);

    if (color_fmt == Black_1_BPP)
    {
        const unsigned int init_size = 1024;
        unsigned int used = 0;
        out->data = (unsigned char*)malloc(init_size);
        out->data_bytes = init_size;

        unsigned int uncompressed_len = (header->cupsWidth + 31) / 32 * 4;
        unsigned char* uncompressed_line_buffer = (unsigned char*)malloc(uncompressed_len);
        memset(uncompressed_line_buffer, 0, uncompressed_len);

        unsigned char* seed = (unsigned char*)malloc(uncompressed_len);
        memset(seed, 0, uncompressed_len);

        unsigned int emit_len = uncompressed_len + 2 + (uncompressed_len + 7) / 8;
        unsigned char* emit = (unsigned char*)malloc(emit_len);

        while (cupsRasterReadPixels(ras, line_buffer, line_len) > 0)
        {
            unsigned char* curr = emit + 2;
            unsigned int diff = 0, len = 0;
            int offset = 0;

            convert_line(line_buffer, line_len, uncompressed_line_buffer);

            do {
                int idx = next_differ(seed, uncompressed_line_buffer, offset, uncompressed_len - offset, &diff);
                if (idx < 0)
                    break;

                if (idx >= 0x1f) {
                    *curr++ = ((diff - 1) << 5) | 0x1f;
                    unsigned int offset_t = idx - 0x1f;
                    while (offset_t >= 0xff) {
                        *curr++ = 0xff;
                        offset_t -= 0xff;
                    }
                    *curr++ = offset_t;
                } else
                    *curr++ = ((diff - 1) << 5) | (idx & 0x1f);

                offset += idx;
                for (int i = 0; i < diff; ++i)
                    *curr++ = uncompressed_line_buffer[offset + i];

                offset += diff;
            } while (offset < uncompressed_len);

            len = curr - emit - 2;
            emit[0] = len;
            emit[1] = len >> 8;
            len += 2;

            if (out->data_bytes < used + len)
            {
                unsigned int new_len = out->data_bytes * 2;
                while (new_len < used + len)
                    new_len *= 2;
                if (!(out->data = (unsigned char*)realloc(out->data, new_len)))
                    break;
                out->data_bytes = new_len;
            }
            memcpy(out->data + used, emit, len);
            used += len;

            unsigned char* temp = seed;
            seed = uncompressed_line_buffer;
            uncompressed_line_buffer = temp;
        }

        out->data_bytes = used;

        free(seed);
        free(uncompressed_line_buffer);
        free(emit);
    }
    else
    {
        struct jpeg_compress_struct cinfo;
        struct jpeg_error_mgr jerr;
        JSAMPROW row_pointer[1];

        cinfo.err = jpeg_std_error(&jerr);
        jpeg_create_compress(&cinfo);

        //jpeg_stdio_dest(&cinfo, stdout);
        jpeg_mem_dest(&cinfo, &out->data, (unsigned long*)&out->data_bytes);

        cinfo.image_width = header->cupsWidth;
        cinfo.image_height = header->cupsHeight;
        cinfo.input_components = 3;
        cinfo.in_color_space = JCS_RGB;

        jpeg_set_defaults(&cinfo);
        //jpeg_set_quality(&cinfo, quality, TRUE);

        jpeg_start_compress(&cinfo, TRUE);

        row_pointer[0] = line_buffer;
        while (cinfo.next_scanline < cinfo.image_height)
        {
            if (cupsRasterReadPixels(ras, line_buffer, line_len) < 1)
            {
                _cupsLangPrintFilter(stderr, "ERROR",
                        _("failed to read raster pixel data:"));
            }
            jpeg_write_scanlines(&cinfo, row_pointer, 1);
        }

        jpeg_finish_compress(&cinfo);
        jpeg_destroy_compress(&cinfo);
    }

    free(line_buffer);

    return out->data_bytes;
}
