#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#include <string.h>
#include <memory.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "cups/cups.h"
#include "cups/ppd.h"
#include "cups/raster.h"
#include "cups/i18n.h"

#include "bud.h"
#include "common.h"
#include "raster_converter.h"

#include <list>
#include <map>
#include <set>

#define DATA_DIR "/opt/unidrv_filter/data"

using namespace std;

static char* g_PaperSize = NULL;
static char* g_Resolution = NULL;
static const char* g_PageMode = NULL;
static const char* g_ColorModel = NULL;
static int g_attributes_count = 0;
static struct tagAttribute* g_attributes = NULL;

typedef const char* CharPtr;
typedef struct tagAttribute* AttributePtr;

struct CharPtrCmp {
    bool operator() ( const CharPtr& x, const CharPtr& y) const
    {
        return strcmp (x, y) < 0;
    }
};

typedef map <CharPtr, AttributePtr, CharPtrCmp> VariableMap;

/*TODO: cache all attributes defined in bud file and all attributes referenced in bud file*/
static VariableMap* g_var_map = NULL;

static int g_features_count = 0;
static int g_commands_count = 0;      /*root level*/
static int g_switch_cases_count = 0;  /*root level*/

/*features + root level commands + root level switch...case*/
static struct tagEntry* g_entries = NULL;

/*TODO: cache each feature's selected option name for fast run switch...case*/
static char** g_features_selected_option_name = NULL;

static struct tagCommandList g_commands[CMD_KIND_COUNT] = {0};

/*helper functions*/
/*NOTE: dbg_log only used in _DEBUG version
 * for log need output in release version, should use _cupsLangPrintFilter and
 * _cupsLangPrintError
 */
#ifdef _DEBUG
static int g_log_level = LOG_LEVEL_DEBUG;

static char* hex2str(unsigned char* data, unsigned int len)
{
    static char hexbuf[256];

    unsigned int i = 0;
    const char* hexstr = "0123456789abcdef";

    if ((uint32_t)len > sizeof(hexbuf)/3) len = sizeof(hexbuf)/3 -1;

    for (i = 0; i<len; ++i)
    {
        hexbuf[3*i+0] = hexstr[data[i]>>4];
        hexbuf[3*i+1] = hexstr[data[i]&0x0f];
        hexbuf[3*i+2] = 0x20;
    }

    hexbuf[3*i] = 0;

    return hexbuf;
}

void dbg_log_imp(int level, const char *fmt, ...)
{
    va_list ap;
    int size = 0;
    char buf[2049];

    if (level < g_log_level)
    {
        return;
    }

    va_start(ap, fmt);
    size += vsnprintf (buf+size, sizeof(buf)-1, fmt, ap);
    va_end(ap);

    _cupsLangPrintFilter(stderr, "DEBUG", "%s", buf);
}
#endif

static char* get_string(FILE* pf, StringTable* table)
{
    unsigned int offset = 0;
    fread(&offset, 1, sizeof(offset), pf);
    if (offset >= table->size)
    {
        dbg_log(LOG_LEVEL_ERR,
                "The string index %08x is invalid, file pos:%08x\n",
                offset, ftell(pf));
        assert(0);
    }

    return table->buf + offset;
}

/*just preorder traversal*/
static void deserialize_ast(struct tagAST** past, FILE* pf, StringTable* table)
{
    char nodetype = 0;
    struct tagAST* ast = NULL;
    fread(&nodetype, 1, sizeof(char), pf);
    switch (nodetype)
    {
    case '+':
    case '-':
    case '*':
    case '/':
        ast = (struct tagAST*)malloc(sizeof(struct tagAST));
        ast->nodetype = nodetype;
        deserialize_ast(&ast->left, pf, table);
        deserialize_ast(&ast->right, pf, table);
        break;

    case 'N':
    case 'R':
        ast = (struct tagAST*)malloc(sizeof(struct tagAST));
        ast->nodetype = nodetype;
        deserialize_ast(&ast->left, pf, table);
        ast->right = NULL;
        break;

    case '%':
    case 'm':
    case 'M':
        ast = (struct tagAST*)malloc(sizeof(struct tagAST));
        ast->nodetype = nodetype;
        deserialize_ast(&ast->left, pf, table);
        deserialize_ast(&ast->right, pf, table);
        break;

    case 'K':
        {
        struct tagNumNode* node = (struct tagNumNode*)malloc(sizeof(struct tagNumNode));
        node->nodetype = nodetype;
        fread(&node->number, 1, sizeof(node->number), pf);
        ast = (struct tagAST*)node;
        }
        break;
    case 'V':
        {
        struct tagVarNode* node = (struct tagVarNode*)malloc(sizeof(struct tagVarNode));
        node->nodetype = nodetype;
        node->name = get_string(pf, table);

        ast = (struct tagAST*)node;
        }
        break;

    default:
        dbg_log(LOG_LEVEL_ERR, "bad node %c\n", nodetype);
        assert(0);
        break;
    }

    *past = ast;
}

static unsigned char char2hexbin(char c)
{
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 0xa;
    else if (c >= 'a' && c <= 'f')
        return c - 'a' + 0xa;
    else
        return c - '0';
}

static void deserialize_cmd(char** cmd, FILE* pf)
{
    unsigned char count = 0;
    fread(&count, 1, 1, pf);
    *cmd = (char*)malloc(count+1);
    **(unsigned char**)cmd = count;
    fread(*cmd+1, 1, count, pf);
}

static void deserialize_arg(struct tagArgList* arg, FILE* pf, StringTable* table)
{
    struct tagCmdArg* item = NULL;
    fread(&arg->count, 1, sizeof(char), pf);
    arg->table = (struct tagCmdArg*)malloc(arg->count *sizeof(struct tagCmdArg));

    item = arg->table;
    for (char i = 0; i<arg->count; ++i)
    {
        fread(&item->type, 1, sizeof(char), pf);
        if ('s' == item->type)
        {
            deserialize_cmd((char**)(&item->ast), pf);
        }
        else
        {
            fread(&item->max, 1, sizeof(item->max), pf);
            fread(&item->min, 1, sizeof(item->min), pf);

            deserialize_ast(&item->ast, pf, table);
            assert (NULL != item->ast);
        }

        ++item;
    }
}

static void deserialize_num_list(struct tagNumList* head, FILE* pf, unsigned int count)
{
    unsigned int i = 0;
    head->type = 'K';
    head->table = (unsigned int*)malloc(count*sizeof(unsigned int));
    head->count = count;
    fread(head->table, sizeof(unsigned int)*count, 1, pf);
}

static void deserialize_str_list(struct tagStrList* head,
        FILE* pf, unsigned int count, StringTable* table)
{
    unsigned int i = 0;
    head->type = 'c';
    head->table = (char**)malloc(count*sizeof(char*));
    head->count = count;

    for(i=0; i<count; ++i)
    {
        head->table[i] = get_string(pf, table);
    }
}

static int deserialize_attribute(struct tagAttribute* attr, FILE* pf, StringTable* table)
{
    unsigned char expr_type = 0;

    fread(&attr->node.type, sizeof(char), 1, pf);
    if (attr->node.type != 'a')
    {
        dbg_log(LOG_LEVEL_ERR,
                "deserialize_attribute error, type:%d, file pos:%08x\n",
                attr->node.type, ftell(pf));
        assert(0);
    }

    attr->type = GPD_NUMBER;
    fread(&attr->type, sizeof(char), 1, pf);

    attr->name = get_string(pf, table);

    dbg_log(LOG_LEVEL_DEBUG, "attribute:%s,%d,%08x\n", attr->name, attr->type, ftell(pf));

    switch (attr->type)
    {
    case GPD_NUMBER:
        fread(&attr->num, sizeof(attr->num), 1, pf);
        break;
    case GPD_STRING:
        attr->str = get_string(pf, table);
        break;
    case GPD_PAIR:
        fread(&attr->pair, sizeof(attr->pair), 1, pf);
        break;
    case GPD_RECT:
        fread(&attr->rect, sizeof(attr->rect), 1, pf);
        break;
    case GPD_LIST:
        {
        char type = 0;
        unsigned int count = 0;
        fread(&type, sizeof(char), 1, pf);
        fread(&count, sizeof(count), 1, pf);
        if ('c' == type)
            deserialize_str_list(&attr->slist, pf, count, table);
        else
            deserialize_num_list(&attr->ilist, pf, count);
        }
        break;
    case GPD_EXPR:
        deserialize_arg(&attr->arg, pf, table);
        break;
    case GPD_MACRO:
       break;
    default:
       dbg_log(LOG_LEVEL_ERR, "%s Not support attribute, type:%02x, file pos:%08x\n",
               __FUNCTION__, attr->type, ftell(pf));
       assert(0);
       break;
    }

    return 0;
}

static int deserialize_entry(struct tagEntry* entry,
        FILE* pf, StringTable* table, struct tagEntryList* parent);

static int deserialize_entry_list(struct tagEntryList** head,
        int count, FILE* pf, StringTable* table, struct tagEntryList* parent)
{
    char sub_type = 0;

    for (int i = 0; i<count; ++i)
    {
        fread(&sub_type, 1, sizeof(char), pf);
        switch (sub_type)
        {
        case 'a':
        {
            struct tagAttribute* it = (struct tagAttribute*)malloc(sizeof(struct tagAttribute));

            fseek(pf, -1, SEEK_CUR);
            deserialize_attribute(it, pf, table);

            head[i] = (struct tagEntryList*)it;
        }
        break;

        case 'e':
        {
            struct tagEntry* it = (struct tagEntry*)malloc(sizeof(struct tagEntry));

            fseek(pf, -1, SEEK_CUR);
            deserialize_entry(it, pf, table, parent);

            head[i] = (struct tagEntryList*)it;
        }
        break;

        default:
        dbg_log(LOG_LEVEL_ERR, "%s [%d/%d]Not support node, type:%02x, file pos:%08x\n",
                __FUNCTION__, i, count, sub_type, ftell(pf));
        assert(0);
        break;
        }
    }

    return 0;
}

static int deserialize_entry(struct tagEntry* entry,
        FILE* pf, StringTable* table, struct tagEntryList* parent)
{
    fread(&entry->node.type, 1, sizeof(char), pf);
    if (entry->node.type != 'e')
    {
        dbg_log(LOG_LEVEL_ERR,
                "deserialize_entry error, type:%d, file pos:%08x\n",
                entry->node.type, ftell(pf));
        assert(0);
    }

    entry->type = ENTRY_FEATURE;
    fread(&entry->type, 1, sizeof(char), pf);

    entry->name = get_string(pf, table);

    fread(&entry->cmd_extral, 1, sizeof(entry->cmd_extral), pf);
    fread(&entry->count, 1, sizeof(entry->count), pf);

    entry->parent = parent;

    dbg_log(LOG_LEVEL_DEBUG, "entry:%s,%d,%d,%08x\n",
            entry->name, entry->type, entry->count, ftell(pf));
    if (entry->count > 0)
    {
        entry->sub = (struct tagEntryList**)malloc(sizeof(struct tagEntryList*)*entry->count);
        deserialize_entry_list(entry->sub, entry->count,
                pf, table, (struct tagEntryList*)entry);
    }

    return 0;
}

static int deserialize_entry_array(struct tagEntry* entry,
        int count, FILE* pf, StringTable* table)
{
    for (int i = 0; i<count; ++i)
    {
        deserialize_entry(entry+i, pf, table, NULL);
    }

    return 0;
}

static int deserialize_root_attributes(struct tagAttribute* head,
         int count, FILE* pf, StringTable* table)
{
    int i = 0;
    struct tagEntryList* it = NULL;
    for (i; i<count; ++i)
    {
        deserialize_attribute(head+i, pf, table);
    }

    return 0;
}

static void align_16bytes(FILE* pf)
{
    char buf[16];
    size_t pos = ftell(pf);
    size_t bytes = pos & 15;
    if (bytes > 0)
    {
        dbg_log(LOG_LEVEL_DEBUG,
                "read %d for align 16bytes, file pos:%08x\n",
                bytes, ftell(pf));
        fread(buf, 1, bytes, pf);
    }
}

static void read_bud_header(FILE* pf, BudFileHeaderV1* header)
{
    fread(&header->id[0], sizeof(header->id), 1, pf);
    fread(&header->version[0], sizeof(header->version), 1, pf);
    fread(&header->author[0], sizeof(header->author), 1, pf);
    fread(&header->time_stamp, sizeof(header->time_stamp), 1, pf);
    fread(&header->file_size, sizeof(header->file_size), 1, pf);

    fread(&header->print_language, sizeof(header->print_language), 1, pf);
    fread(&header->color_space, sizeof(header->color_space), 1, pf);
    fread(&header->bits_per_pixel, sizeof(header->bits_per_pixel), 1, pf);
    fread(&header->sub_blocks, sizeof(header->sub_blocks), 1, pf);

    fread(&header->string_table, sizeof(header->string_table), 1, pf);
    fread(&header->attributes, sizeof(header->attributes), 1, pf);
    fread(&header->commands, sizeof(header->commands), 1, pf);
    fread(&header->features, sizeof(header->features), 1, pf);
    fread(&header->switch_cases, sizeof(header->switch_cases), 1, pf);
}

static void cache_attribute(struct tagAttribute* attr)
{
    VariableMap::iterator it = g_var_map->lower_bound(attr->name);
    if (it != g_var_map->end() && !g_var_map->key_comp()(attr->name, (*it).first))
    {
        if (GPD_STRING == attr->type)
            dbg_log(LOG_LEVEL_DEBUG, "attribute %s:%s conflict with cached name\n",
                attr->name, attr->str);
        else
            dbg_log(LOG_LEVEL_DEBUG, "attribute %s:%d conflict with cached name\n",
                attr->name, attr->type);
        return;
    }

    g_var_map->insert(it, pair<CharPtr, AttributePtr>(attr->name, attr));
}

static struct tagAttribute* get_attribute( const char* name)
{
    VariableMap::iterator it = g_var_map->find(name);

    if (it != g_var_map->end())
        return (*it).second;

    return NULL;
}

static int load_bud_file(char* filename, BudFileHeaderV1* header)
{
    int i = 0;
    FILE* pf = NULL;
    StringTable table = {NULL,0,0};

    pf = fopen(filename, "rb");
    if (NULL == pf)
    {
        return -1;
    }

    dbg_log(LOG_LEVEL_INFO, "start de-serialize from %s...\n", filename);

    read_bud_header(pf, header);
    assert(header->sub_blocks < 10);

    if (memcmp(header->id, "budfake", sizeof(header->id)-1))
    {
        fclose(pf);
        return -2;
    }

    g_attributes_count = header->attributes.count;
    g_features_count = header->features.count;
    g_commands_count = header->commands.count;
    g_switch_cases_count = header->switch_cases.count;
    g_entries = (struct tagEntry*)malloc(
            (g_features_count+g_commands_count+g_switch_cases_count)*
            sizeof(struct tagEntry));

    /*0, load string table here!*/
    table.buf = (char*)malloc(header->string_table.count+1);
    fseek(pf, header->string_table.offset, SEEK_SET);
    fread(table.buf, 1, header->string_table.count, pf);
    table.size = header->string_table.count;
    table.buf[table.size] = 0;
    dbg_log(LOG_LEVEL_INFO, "load string table:%08x,%08x\n",
            header->string_table.offset,
            header->string_table.count);

    /*1, de-serialize all root level attributes*/
    g_attributes = (struct tagAttribute*)malloc(g_attributes_count*sizeof(struct tagAttribute));
    fseek(pf, header->attributes.offset, SEEK_SET);
    dbg_log(LOG_LEVEL_INFO, "load root level attributes:%08x,%d, file pos:%08x\n",
            header->attributes.offset,
            header->attributes.count, ftell(pf));
    deserialize_root_attributes(g_attributes, g_attributes_count, pf, &table);
    align_16bytes(pf);

    /*2, de-serialize all features*/
    dbg_log(LOG_LEVEL_INFO, "load all features :%08x,%d, file pos:%08x\n",
            header->features.offset,
            header->features.count, ftell(pf));
    deserialize_entry_array(g_entries, g_features_count, pf, &table);
    align_16bytes(pf);

    /*3, de-serialize all root level commands*/
    dbg_log(LOG_LEVEL_INFO, "load root level commands :%08x,%d, file pos:%08x\n",
            header->commands.offset,
            header->commands.count, ftell(pf));
    deserialize_entry_array(g_entries+g_features_count, g_commands_count, pf, &table);
    align_16bytes(pf);

    /*4, de-serialize all root level switch...case*/
    dbg_log(LOG_LEVEL_INFO, "load root level switch...case :%08x,%d, file pos:%08x\n",
            header->switch_cases.offset,
            header->switch_cases.count, ftell(pf));
    deserialize_entry_array(g_entries+g_features_count+g_commands_count,
            g_switch_cases_count, pf, &table);
    align_16bytes(pf);

    fclose(pf);

#ifdef _DEBUG
    if (g_log_level < LOG_LEVEL_WARN)
    {
        print_attributes(g_attributes, g_attributes_count);

        print_entries(g_entries,
                g_features_count+g_commands_count+g_switch_cases_count);
    }
#endif

    dbg_log(LOG_LEVEL_INFO, "finish de-serialize from %s..., sub_blocks=%d\n",
            filename, header->sub_blocks);

    return 0;
}

static void init_command_lists(void)
{
    for (unsigned int i = 0; i<CMD_KIND_COUNT; ++i)
    {
        list_init(&g_commands[i].entry);
    }
}

static void free_command_lists(void)
{
    for (unsigned int i = 0; i<CMD_KIND_COUNT; ++i)
    {
        struct tagCommandList *it, *next;

        LIST_FOR_EACH_ENTRY_SAFE( it, next, &g_commands[i].entry, struct tagCommandList, entry )
        {
            list_remove(&it->entry);
            free(it);
        }
    }
}

static void push_command_entry(struct tagEntry* entry)
{
    unsigned index = GET_CMD_KIND(entry->cmd_extral);

    struct tagCommandList* item = (struct tagCommandList*)malloc(sizeof(struct tagCommandList));
    list_init(&item->entry);
    item->cmd = entry;

    if (index <= CMD_JOB_FINISH)
    {
        /*sort command by section order*/
        struct tagCommandList* it = NULL;
        LIST_FOR_EACH_ENTRY(it, &g_commands[index].entry, struct tagCommandList, entry)
        {
            struct tagEntry* ptr = it->cmd;
            if (GET_CMD_ORDER(ptr->cmd_extral) > GET_CMD_ORDER(entry->cmd_extral))
            {
                break;
            }
        }

        list_add_tail(&it->entry, &item->entry);
    }
    else
    {
        list_add_tail(&g_commands[index].entry, &item->entry);
    }
}

static char* get_feature_attrib_imp (struct tagEntryList** head,
        unsigned int count, const char* option_name)
{
    struct tagEntryList* it = NULL;

    for (unsigned int i = 0; i<count; ++i)
    {
        it = head[i];
        switch (it->type)
        {
        case 'a':
            if (!strcmp(option_name, ((struct tagAttribute*)it)->name))
            {
                return ((struct tagAttribute*)it)->str;
            }
        break;

        case 'e':
        break;

        case 'p':
        break;

        default:
        assert(0);
        break;
        }
    }

    return NULL;
}

static char* get_feature_attrib(char* feature_name, const char* option_name)
{
    for (int i = 0; i<g_features_count; ++i)
    {
        struct tagEntry* entry = &g_entries[i];

        if (!strcmp(feature_name, entry->name))
        {
            char* value = get_feature_attrib_imp(entry->sub, entry->count, option_name);
            if (value) return value;
        }
    }

    return NULL;
}

static void execute_gpd_imp(struct tagEntryList** head, unsigned int count, int level)
{
    for (unsigned int i = 0; i<count; ++i)
    {
        struct tagEntryList* it = head[i];

        switch (it->type)
        {
        case 'a':
            {
                char* name = ((struct tagAttribute*)it)->name;
                if (strcmp("DefaultOption", name) && strcmp("Name", name))
                {
                    cache_attribute((struct tagAttribute*)it);
                }
            }
        break;

        case 'e':
        {
            int parse_sub = 1;
            struct tagEntry* entry = (struct tagEntry*)it;

            switch (entry->type)
            {
            case ENTRY_FEATURE:
                break;
            case ENTRY_OPTION:
                {
                    const char* value = NULL;
                    struct tagEntry* parent = (struct tagEntry*)entry->parent;

                    if (!strcmp(parent->name, "PaperSize") && g_PaperSize)
                    {
                        value = g_PaperSize;
                    }
                    else if (!strcmp(parent->name, "Resolution") && g_Resolution)
                    {
                        value = g_Resolution;
                    }
                    else if (!strcmp(parent->name, "Duplex") && g_PageMode)
                    {
                        value = g_PageMode;
                    }
                    else if (!strcmp(parent->name, "ColorModel") && g_ColorModel)
                    {
                        value = g_ColorModel;
                    }
                    else
                    {
                        value = get_feature_attrib_imp(parent->sub,
                                                       parent->count,
                                                       "DefaultOption");
                    }

                    if (NULL != value) /*may be no DefaultOption attribute*/
                    {
                        parse_sub = !strcmp(entry->name, value);
                        if (parse_sub)
                        {
                            dbg_log(LOG_LEVEL_DEBUG,
                                "required %s, got %s\n", value, parent->name);
                        }
                    }
                }

                break;
            case ENTRY_COMMAND:
                break;
            case ENTRY_SWITCH:
                {
                    char* switch_value = NULL;
                    if (!strcmp(entry->name, "PaperSize") && g_PaperSize)
                    {
                        switch_value = g_PaperSize;
                    }
                    else if (!strcmp(entry->name, "Resolution") && g_Resolution)
                    {
                        switch_value = g_Resolution;
                    }
                    else
                    {
                        switch_value = get_feature_attrib(entry->name, "DefaultOption");
                    }
                    assert(switch_value);

                    dbg_log(LOG_LEVEL_DEBUG, "replace switch %s -> switch %s\n",
                            entry->name, switch_value);
                    entry->name = switch_value;
                }
                break;
            case ENTRY_CASE:
                {
                    struct tagEntry* parent = (struct tagEntry*)entry->parent;
                    assert(parent);
                    assert(parent->name);
                    assert(parent->type == ENTRY_SWITCH);
                    parse_sub = !strcmp(entry->name, parent->name);
                }
                break;
            case ENTRY_FONTCAR:
                break;
            default:
                assert(0);
               break;
            }

            if (ENTRY_COMMAND == entry->type)
            {
                push_command_entry(entry);
            }
            else if (entry->sub && parse_sub)
            {
                execute_gpd_imp(entry->sub, entry->count, level + 1);
            }
        }
        break;

        default:
        assert(0);
        break;
        }
    }
}

static void guess_resolution(ppd_file_t* ppd, unsigned int width, unsigned int height)
{
    char cur_resolution[100];
    snprintf(cur_resolution, sizeof(cur_resolution), "%dx%ddpi", width, height);
    ppd_option_t* op = ppdFindOption(ppd, "Resolution");
    if (op)
    {
        for (int i = 0; i<op->num_choices; ++i)
        {
            if (!strcmp(op->choices[i].choice, cur_resolution))
            {
                if (op->choices[i].text[0] > 0x20)
                {
                    g_Resolution = strdup(op->choices[i].text);
                    break;
                }
            }
        }
    }
}

/*Got all command, sort by command kind
@filter_options　example:
argv[5]="PageSize=A4 Collate ColorModel=Grayscale Duplex=None job-uuid=urn:uuid:4efb51fc-2892-349b-7268-8c628bd1e260 job-originating-host-name=localhost date-time-at-creation= date-time-at-processing= time-at-creation=1486964168 time-at-processing=1486964168"
argv[5]="number-up=1 PageSize=A4 noCollate job-uuid=urn:uuid:4062fa75-2fb4-3022-77b8-05e4912b3853 job-originating-host-name=localhost date-time-at-creation= date-time-at-processing= time-at-creation=1486968307 time-at-processing=1486968307"

some times: argv[5] have Resolution=wxhdpi:
argv[5]="PageSize=A4 number-up=1 noCollate Resolution=180x180dpi job-uuid=urn:uuid:72f5bc90-6412-331a-4b88-3fe0fba766ef job-originating-host-name=localhost date-time-at-creation= date-time-at-processing= time-at-creation=1490851197 time-at-processing=1490851197"

*/
static int execute_gpd(const char* filter_options)
{
    int level = 0;
    struct tagEntry* features = g_entries;
    int count = g_features_count+g_commands_count+g_switch_cases_count;
    const char* pos = strstr(filter_options, "PageSize=");
    if (pos)
    {
        pos += strlen("PageSize=");

        int len = 0;
        const char* pos2 = strchr(pos, ' ');
        if (pos2)
            len = int(pos2 - pos);
        else
            len = strlen(pos);
        if (len > 1)
        {
            g_PaperSize = (char*)malloc(len+1);
            memcpy(g_PaperSize, pos, len);
            g_PaperSize[len] = 0;
        }
    }

    dbg_log(LOG_LEVEL_DEBUG, "%s:%p,%p,%d,%s,%s\n",
            __FUNCTION__, filter_options, features, count,
            g_PaperSize, g_Resolution);

    g_var_map = new VariableMap;
    init_command_lists();

    for (int i = 0; i<g_attributes_count; ++i)
    {
        cache_attribute(&g_attributes[i]);
    }

    for (int i = 0; i<count; ++i)
    {
        struct tagEntryList* ptr = (struct tagEntryList*)&features[i];
        execute_gpd_imp(&ptr, 1, 0);
    }

    const char* cursor_attr_names[4] = {
        "DestX",
        "DestXRel",
        "DestY",
        "DestYRel",
    };
    for (int i = 0; i<sizeof(cursor_attr_names)/sizeof(cursor_attr_names[0]); ++i)
    {
        struct tagAttribute* attr = get_attribute(cursor_attr_names[i]);
        if (NULL == attr)
        {
            attr = (struct tagAttribute*)malloc(sizeof(*attr));
            attr->node.type = 'a';
            attr->type = GPD_NUMBER;
            attr->name = (char*)cursor_attr_names[i];
            attr->num = 0;
            cache_attribute(attr);
        }
    }

#ifdef _DEBUG
    if (g_log_level < LOG_LEVEL_WARN)
    {
        for (int i = 0; i<CMD_KIND_COUNT; ++i)
        {
            struct tagCommandList* it = NULL;
            LIST_FOR_EACH_ENTRY(it, &g_commands[i].entry, struct tagCommandList, entry)
            {
                print_entries(it->cmd, 1);
            }
        }
    }
#endif

    return 0;
}

static int eval_ast(struct tagAST* ast, void* params)
{
    switch (ast->nodetype)
    {
    case '+':
        {
            return eval_ast(ast->left, params) +
                   eval_ast(ast->right, params);
        }
        break;

    case '-':
        {
            return eval_ast(ast->left, params) -
                   eval_ast(ast->right, params);
        }
        break;

    case '*':
        {
            return eval_ast(ast->left, params) *
                   eval_ast(ast->right, params);
        }
        break;

    case '/':
        {
            return eval_ast(ast->left, params) /
                   eval_ast(ast->right, params);
        }
        break;

    case 'N':
        {
            int value = eval_ast(ast->left, params);
            assert (value>0);
            return -value;
        }
        break;

    case 'R':
        {
            *(char*)params = 1;
            return eval_ast(ast->left, params);
        }
        break;

    case '%':
        {
            return eval_ast(ast->left, params) %
                   eval_ast(ast->right, params);
        }
        break;

    case 'm':
        {
            int v1 = eval_ast(ast->left, params);
            int v2 = eval_ast(ast->right, params);
            return v1<v2 ? v1:v2;
        }
        break;

    case 'M':
        {
            int v1 = eval_ast(ast->left, params);
            int v2 = eval_ast(ast->right, params);
            return v1<v2 ? v1:v2;
        }
        break;

    case 'K':
        {
            return ((struct tagNumNode*)ast)->number;
        }
        break;
    case 'V':
        {
            struct tagVarNode* var = (struct tagVarNode*)ast;
            struct tagAttribute* attr = get_attribute(var->name);
            if (NULL == attr)
            {
                attr = (struct tagAttribute*)malloc(sizeof(*attr));
                attr->node.type = 'a';
                attr->type = GPD_NUMBER;
                attr->name = var->name;
                attr->num = 0;
                cache_attribute(attr);
            }

            assert(GPD_NUMBER == attr->type);
            return attr->num;
        }
        break;

    default:
        break;
    }

    return 0;
}

static int format_cmd_arg(char type, int value,
        unsigned char* out_buf, unsigned int out_buf_bytes )
{
    unsigned char* walk = out_buf;
    switch (type)
    {
    case 'c':
        *walk++ = value;
        break;
    case 'd':
        walk += sprintf((char*)walk, "%d", value);
        break;
    case 'x':
        *(unsigned int*)walk = (unsigned int)value;
        walk += sizeof(unsigned int);
        break;
    case 'l':
        *walk++ = value;
        *walk++ = (value>>8);
        break;
    case 'm':
        *walk++ = value>>8;
        *walk++ = value;
        break;
    default:
        _cupsLangPrintFilter(stderr, "ERROR",
                _("Unsupported Argument Type Specifier:%c"), type);
        break;
    }

    return int(walk - out_buf);
}

static void write_printer(unsigned char* buf, unsigned int size)
{
    fwrite(buf, 1, size, stdout);

#ifdef _DEBUG
    unsigned int bytes = (size>48) ? 48:size;
    _cupsLangPrintFilter(stderr, "DEBUG",
            _("Write to printer %d:%s"),
            size, hex2str(buf, bytes));
#endif
}

static int execute_cmd_expr(struct tagEntry* cmd, void* params)
{
    unsigned char hexbuf[256];
    struct tagAttribute* attr = NULL;

    if (1)
    {
        attr = (struct tagAttribute*)(cmd->sub[0]);
        assert(!strcmp(attr->name, "Cmd"));
    }
    else
    {
        for (unsigned int i = 0; i<cmd->count; ++i)
        {
            attr = (struct tagAttribute*)(cmd->sub[i]);
            assert('a' == attr->node.type);

            if (!strcmp(attr->name, "Cmd"))
            {
                break;
            }
        }
    }

    struct tagCmdArg* arg = attr->arg.table;
    for (char i = 0; i<attr->arg.count; ++i)
    {
        if ('s' == arg->type)
        {
            unsigned char size = *(unsigned char*)(arg->ast);
            write_printer((unsigned char*)(arg->ast)+1, size);
        }
        else
        {
            int cmd_len = 0;
            char max_repeate = 0;
            int value = eval_ast(arg->ast, &max_repeate);

            if (max_repeate)
            {
                while (value >= arg->max)
                {
                    value -= arg->max;
                    cmd_len = format_cmd_arg(arg->type, arg->max,
                                             hexbuf, sizeof(hexbuf));
                    write_printer(hexbuf, cmd_len);
                }

                if (value >= arg->min)
                {
                    cmd_len = format_cmd_arg(arg->type, value,
                                             hexbuf, sizeof(hexbuf));
                    write_printer(hexbuf, cmd_len);
                }
            }
            else
            {
                if (arg->min < arg->max)
                {
                    if (value > arg->max) value = arg->max;
                    if (value < arg->min) value = arg->min;
                }
                cmd_len = format_cmd_arg(arg->type, value,
                                         hexbuf, sizeof(hexbuf));
                write_printer(hexbuf, cmd_len);
            }
        }

        ++arg;
    }

    return 0;
}

static struct tagEntry* get_cmd_by_name(int index, const char* cmd_name)
{
    struct tagCommandList* it = NULL;
    struct tagEntry* cmd = NULL;
    LIST_FOR_EACH_ENTRY(it, &g_commands[index].entry, struct tagCommandList, entry)
    {
        if (!strcmp(it->cmd->name, cmd_name))
        {
            return it->cmd;
        }
    }

    return NULL;
}

/* Has the current job been canceled? */
static int		g_Canceled = 0;

/*
 * 'Setup()' - Prepare the printer for printing.
 */

static void Setup(ppd_file_t* ppd, int* block_lines, bool* interlacing,
        bool* h_byte, struct tagPair* dpi, struct tagPair* master_unit)
{
    struct tagAttribute* attr = get_attribute("OutputDataFormat");
    if (attr)
    {
        struct tagAttribute* attr2 = get_attribute("PinsPerPhysPass");
        struct tagAttribute* attr3 = get_attribute("PinsPerLogPass");
        assert(NULL != attr2);
        assert(NULL != attr3);

        if (!strcmp(attr->str, "V_BYTE"))
        {
            *h_byte = false;
            *block_lines = attr2->num;
            if (attr2->num*2 == attr3->num)
            {
                /* Sometime interleaving is same as shingling.
                 * "shingling" has a history as a term for both Epson and HP printers.*/
                *interlacing = true;
            }
        }
        else
        {
            *h_byte = true;
            *block_lines = attr3->num;
        }
    }
    else
    {
        *block_lines = 1; /*FIXME: is it right if no OutputDataFormat */
    }

    attr = get_attribute("DPI");
    assert(attr != NULL);
    memcpy(dpi, &attr->pair, sizeof(*dpi));

    attr = get_attribute("MasterUnits");
    assert(attr != NULL);
    memcpy(master_unit, &attr->pair, sizeof(*master_unit));

    struct tagCommandList* it = NULL;
    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_JOB_SETUP].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }

    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_DOC_SETUP].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }
}


/*
 * 'StartPage()' - Start a page of graphics.
 */

static void
StartPage(
    const ppd_file_t         *ppd,	    /* I - PPD file */
    const cups_page_header2_t *header)	/* I - Page header */
{

    /*
    * Show page device dictionary...
    */

    fprintf(stderr, "DEBUG: StartPage...\n");
    fprintf(stderr, "DEBUG: Duplex = %d\n", header->Duplex);
    fprintf(stderr, "DEBUG: HWResolution = [ %d %d ]\n", header->HWResolution[0], header->HWResolution[1]);
    fprintf(stderr, "DEBUG: ImagingBoundingBox = [ %d %d %d %d ]\n", header->ImagingBoundingBox[0], header->ImagingBoundingBox[1], header->ImagingBoundingBox[2], header->ImagingBoundingBox[3]);
    fprintf(stderr, "DEBUG: Margins = [ %d %d ]\n", header->Margins[0], header->Margins[1]);
    fprintf(stderr, "DEBUG: ManualFeed = %d\n", header->ManualFeed);
    fprintf(stderr, "DEBUG: MediaPosition = %d\n", header->MediaPosition);
    fprintf(stderr, "DEBUG: NumCopies = %d\n", header->NumCopies);
    fprintf(stderr, "DEBUG: Orientation = %d\n", header->Orientation);
    fprintf(stderr, "DEBUG: PageSize = [ %d %d ]\n", header->PageSize[0], header->PageSize[1]);
    fprintf(stderr, "DEBUG: cupsWidth = %d\n", header->cupsWidth);
    fprintf(stderr, "DEBUG: cupsHeight = %d\n", header->cupsHeight);
    fprintf(stderr, "DEBUG: cupsMediaType = %d\n", header->cupsMediaType);
    fprintf(stderr, "DEBUG: cupsBitsPerColor = %d\n", header->cupsBitsPerColor);
    fprintf(stderr, "DEBUG: cupsBitsPerPixel = %d\n", header->cupsBitsPerPixel);
    fprintf(stderr, "DEBUG: cupsBytesPerLine = %d\n", header->cupsBytesPerLine);
    fprintf(stderr, "DEBUG: cupsColorOrder = %d\n", header->cupsColorOrder);
    fprintf(stderr, "DEBUG: cupsColorSpace = %d\n", header->cupsColorSpace);
    fprintf(stderr, "DEBUG: cupsCompression = %d\n", header->cupsCompression);

    struct tagCommandList* it = NULL;
    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_PAGE_SETUP].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }
}


/*
 * 'EndPage()' - Finish a page of graphics.
 */

static void
EndPage(
    const cups_page_header2_t *header)	/* I - Page header */
{
    /*
    * Eject the current page...
    */
    struct tagCommandList* it = NULL;
    struct tagAttribute* attr = get_attribute("EjectPageWithFF");
    if ( attr && attr->num>0 )
    {
        LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_CURSOR_MOVE].entry, struct tagCommandList, entry)
        {
            if (!strcmp(it->cmd->name, "CmdFF"))
                execute_cmd_expr(it->cmd, NULL);
        }
    }

    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_PAGE_FINISH].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }

    fflush(stdout);
}


/*
 * 'Shutdown()' - Shutdown the printer.
 */

static void
Shutdown(void)
{
    struct tagCommandList* it = NULL;
    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_DOC_FINISH].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }

    LIST_FOR_EACH_ENTRY(it, &g_commands[CMD_JOB_FINISH].entry, struct tagCommandList, entry)
    {
        execute_cmd_expr(it->cmd, NULL);
    }
}


/*
 * 'CancelJob()' - Cancel the current job...
 */

static void
CancelJob(int sig)			/* I - Signal */
{
    (void)sig;

    g_Canceled = 1;
}

static PrintLanguage get_print_langugae (const char* language)
{
    if (!strcmp(language, "ESC/P"))
    {
        return PL_ESC_P;
    }
    else if (!strcmp(language, "ESC/P2"))
    {
        return PL_ESC_P_2;
    }
    else if (!strcmp(language, "PCL5"))
    {
        return PL_PCL_5;
    }
    else if (!strcmp(language, "PCLXL"))
    {
        return PL_PCL_XL;
    }
    else if (!strcmp(language, "UFR"))
    {
        return PL_UFR;
    }
    else if (!strcmp(language, "UFRII"))
    {
        return PL_UFR_II;
    }
    else if (!strcmp(language, "BRScript2"))
    {
        return PL_BR_SCRIPT_2;
    }
    else if (!strcmp(language, "BRScript3"))
    {
        return PL_BR_SCRIPT_3;
    }
    else if (!strcmp(language, "EPSONScript"))
    {
        return PL_EPSON_SCRIPT;
    }

    return PL_UNKNOWN;
}

/*
 * 'main()' - Main entry and processing of driver.
 */
int					            /* O - Exit status */
main(int  argc,				    /* I - Number of command-line arguments */
     char *argv[])			    /* I - Command-line arguments */
{
    BudFileHeaderV1 bud_header;
    int	fd;		                /* File descriptor */
    cups_raster_t *ras;         /* Raster stream for printing */
    cups_page_header2_t	header;	/* Page header from file */
    ppd_file_t *ppd;		    /* PPD file */
    int	page;		            /* Current page */
    int ret = 0;
    bool h_byte = true;
    bool interlacing = false;
    int block_lines = 0;
    int block_count = 0;
    RasterBlock* raster_block = NULL;
    ColorFmt color_fmt;
    PrintLanguage language;
    int last_x = 0;
    int last_y = 0;
    int margin_x = 0;
    int margin_y = 0;
    struct tagPair current_dpi;
    struct tagPair master_unit;

    struct tagEntry* CmdCR = NULL;
    struct tagEntry* CmdYMoveAbsolute = NULL;
    struct tagEntry* CmdYMoveRelDown = NULL;
    struct tagEntry* CmdXMoveAbsolute = NULL;
    struct tagEntry* CmdXMoveRelRight = NULL;
    struct tagEntry* CmdXMoveRelLeft = NULL;
    struct tagEntry* CmdSendBlockData = NULL;
    struct tagEntry* CmdBeginRaster = NULL;
    struct tagEntry* CmdEndRaster = NULL;
    struct tagAttribute* y_attr = NULL;
    struct tagAttribute* x_attr = NULL;
    struct tagAttribute* NumOfDataBytes = NULL;
    struct tagAttribute* NumOfCopies = NULL;
    struct tagAttribute* RasterDataWidthInPixels = NULL;
    struct tagAttribute* RasterDataHeightInPixels = NULL;
    struct tagAttribute* PageNumber = NULL;


#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
    struct sigaction action;		/* Actions for POSIX signals */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

    init_log();

#ifdef _DEBUG
    if (argc > 1)
    {
        if (!strcmp(argv[1], "test"))
        {
            g_log_level = atoi(argv[3]);
            load_bud_file(argv[2], &bud_header);
            execute_gpd(argv[4]);

            uninit_log();
            if (g_var_map) delete g_var_map;
            return 1;
        }
        else if (!strcmp(argv[1], "test2"))
        {
            block_lines = 24;
            bud_header.sub_blocks = 3;
            h_byte = false;
            fd = open(argv[2], O_RDONLY);
            ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
            cupsRasterReadHeader2(ras, &header);

            interlacing = atoi(argv[3])>0;
            fprintf (stderr, "fd=%d, ras=%p, ret=%d\n",  fd, ras,
                convert_raster_block(ras,
                &header,
                Black_1_BPP,
                h_byte,
                interlacing,
                block_lines,
                bud_header.sub_blocks,
                &raster_block,
                &block_count));
            close(fd);

            uninit_log();
            if (g_var_map) delete g_var_map;
            return 1;
        }
        else if (!strcmp(argv[1], "test3"))
        {
            char* ppd_file = getenv("PPD");
            ppd = ppdOpenFile(ppd_file);
            if (ppd) guess_resolution(ppd, atoi(argv[1]), atoi(argv[2]));
            uninit_log();
            if (g_var_map) delete g_var_map;
            return 1;
        }
        else if (!strcmp(argv[1], "pclxl"))
        {
            fd = 0;
            ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
            cupsRasterReadHeader2(ras, &header);

            fprintf (stderr, "fd=%d, ras=%p, ret=%d\n",  fd, ras,
                ret = (convert_raster_block_pclxl(ras,
                &header,
                &raster_block,
                &block_count)));

            fwrite(raster_block[0].data, 1, ret, stdout);

            close(fd);

            uninit_log();
            if (g_var_map) delete g_var_map;
            return 1;
        }
    }
#endif  /*end  #ifdef _DEBUG*/

    /*
    * Make sure status messages are not buffered...
    */
    setbuf(stderr, NULL);

    /*
    * Check command-line...
    */
    if (argc < 6 || argc > 7)
    {
        /*
        * We don't have the correct number of arguments; write an error message
        * and return.
        */

        _cupsLangPrintFilter(stderr, "ERROR",
                             _("%s job-id user title copies options [file]"),
                             argv[0]);
        uninit_log();
        if (g_var_map) delete g_var_map;
        return (1);
    }

    /*
    * Open the page stream...
    */
    if (argc == 7)
    {
        if ((fd = open(argv[6], O_RDONLY)) == -1)
        {
            _cupsLangPrintError("ERROR", _("Unable to open raster file"));
            sleep(1);
            uninit_log();
            if (g_var_map) delete g_var_map;
            return (1);
        }
    }
    else
    {
        fd = 0;
    }

    ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
    if (!ras || !cupsRasterReadHeader2(ras, &header))
    {
        _cupsLangPrintError("ERROR", _("Unable to read raster file"));
        sleep(1);
        uninit_log();
        if (g_var_map) delete g_var_map;
        return (1);
    }

    /*
    * Register a signal handler to eject the current page if the
    * job is cancelled.
    */
    g_Canceled = 0;

#ifdef HAVE_SIGSET /* Use System V signals over POSIX to avoid bugs */
    sigset(SIGTERM, CancelJob);
#elif defined(HAVE_SIGACTION)
    memset(&action, 0, sizeof(action));

    sigemptyset(&action.sa_mask);
    action.sa_handler = CancelJob;
    sigaction(SIGTERM, &action, NULL);
#else
    signal(SIGTERM, CancelJob);
#endif /* HAVE_SIGSET */

    /*
    * Initialize the print device...
    */
    char* ppd_file = getenv("PPD");
    ppd = ppdOpenFile(ppd_file);
    if (!ppd)
    {
        ppd_status_t	status;	/* PPD error */
        int			linenum;	/* Line number */

        _cupsLangPrintFilter(stderr, "ERROR",
                             _("The PPD file could not be opened."));

        status = ppdLastError(&linenum);

        fprintf(stderr, "DEBUG: %s on line %d.\n", ppdErrorString(status), linenum);

        uninit_log();
        if (g_var_map) delete g_var_map;
        return (1);
    }

    /*
    ppd_file_t->manufacturer; Epson
    ppd_file_t->modelname;    Epson 24-Pin Series
    ppd_file_t->nickname;     Epson 24-Pin Series
    ppd_file_t->pcfilename;   epson24.ppd
    ppd_file_t->product;      (24-Pin Series)
    argv[0]="NFCP-DPK890H"
    envp[20]="PPD=/etc/cups/ppd/NFCP-DPK890H.ppd"
    envp[23]="DEVICE_URI=usb://NFCP/DPK890H
    envp[26]="PRINTER=NFCP-DPK890H
    */
    {
        char bud_filename[256];
        int len = strlen(ppd->pcfilename);
        ppd->pcfilename[len-3] = 'b';
        ppd->pcfilename[len-2] = 'u';
        ppd->pcfilename[len-1] = 'd';
        ppd->pcfilename[len] = 0;

        snprintf(bud_filename, sizeof(bud_filename), DATA_DIR"/%s", ppd->pcfilename);
        ret = load_bud_file(bud_filename, &bud_header);
        if (ret < 0)
        {
            _cupsLangPrintFilter(stderr, "ERROR",
                                 _("The bud file %s could not be opened (%d)."),
                                 bud_filename, ret);
            uninit_log();
            return (1);
        }

        guess_resolution(ppd, header.HWResolution[0], header.HWResolution[1]);

        if (header.Duplex)
        {
            if(header.Tumble)
                g_PageMode = "HORIZONTAL";
            else
                g_PageMode = "VERTICAL";
        }

        if (header.cupsColorSpace == CUPS_CSPACE_RGB)
            g_ColorModel = "RGB";
        else if (header.cupsColorSpace == CUPS_CSPACE_K)
            g_ColorModel = "Gray";

        ret = execute_gpd(argv[5]);
        if (ret < 0)
        {
            _cupsLangPrintFilter(stderr, "ERROR",
                                 _("The bud file %s could not be executed (%d)."),
                                ppd->pcfilename, ret);
            uninit_log();
            if (g_var_map) delete g_var_map;
            return (1);
        }
    }

    Setup(ppd, &block_lines, &interlacing, &h_byte, &current_dpi, &master_unit);
    if (block_lines % bud_header.sub_blocks)
    {
        bud_header.sub_blocks = 1;
    }
    _cupsLangPrintFilter(stderr, "DEBUG",
            _("(%dx%d)/(%dx%d), %s, %s, %dx%d, i=%d, h=%d"),
            current_dpi.first, current_dpi.second,
            master_unit.first, master_unit.second,
            g_PaperSize, g_Resolution,
            header.HWResolution[0], header.HWResolution[1],
            interlacing, h_byte);

    language = get_print_langugae(bud_header.print_language);
    if (PL_UNKNOWN == language)
    {
        _cupsLangPrintFilter(stderr, "ERROR",
                             _("Unknown print language:%s"), bud_header.print_language);
        uninit_log();
        if (g_var_map) delete g_var_map;
        return (1);
    }

    CmdCR = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdCR");

    CmdYMoveAbsolute = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdYMoveAbsolute");
    if (CmdYMoveAbsolute)
    {
        y_attr = get_attribute("DestY");
    }
    else
    {
        y_attr = get_attribute("DestYRel");
        CmdYMoveRelDown = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdYMoveRelDown");
    }

    CmdXMoveAbsolute = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdXMoveAbsolute");
    if (CmdXMoveAbsolute)
    {
        x_attr = get_attribute("DestX");
    }
    else
    {
        x_attr = get_attribute("DestXRel");
        CmdXMoveRelRight = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdXMoveRelRight");
        CmdXMoveRelLeft = get_cmd_by_name(CMD_CURSOR_MOVE, "CmdXMoveRelLeft");
    }

    CmdSendBlockData = get_cmd_by_name(CMD_RASTER_SEND, "CmdSendBlockData");
    CmdBeginRaster = get_cmd_by_name(CMD_GENERIC_SET, "CmdBeginRaster");
    CmdEndRaster = get_cmd_by_name(CMD_GENERIC_SET, "CmdEndRaster");
    NumOfDataBytes = get_attribute("NumOfDataBytes");
    if (NULL == NumOfDataBytes)
    {
        NumOfDataBytes = (struct tagAttribute*)malloc(sizeof(*NumOfDataBytes));
        NumOfDataBytes->node.type = 'a';
        NumOfDataBytes->type = GPD_NUMBER;
        NumOfDataBytes->name = (char*)"NumOfDataBytes";
        cache_attribute(NumOfDataBytes);
    }

    NumOfCopies = get_attribute("NumOfCopies");
    if (NULL == NumOfCopies)
    {
        NumOfCopies = (struct tagAttribute*)malloc(sizeof(*NumOfCopies));
        NumOfCopies->node.type = 'a';
        NumOfCopies->type = GPD_NUMBER;
        NumOfCopies->name = (char*)"NumOfCopies";
        cache_attribute(NumOfCopies);
    }
    NumOfCopies->num = header.NumCopies;

    RasterDataWidthInPixels = get_attribute("RasterDataWidthInPixels");
    if (NULL == RasterDataWidthInPixels)
    {
        RasterDataWidthInPixels = (struct tagAttribute*)malloc(sizeof(*RasterDataWidthInPixels));
        RasterDataWidthInPixels->node.type = 'a';
        RasterDataWidthInPixels->type = GPD_NUMBER;
        RasterDataWidthInPixels->name = (char*)"RasterDataWidthInPixels";
        cache_attribute(RasterDataWidthInPixels);
    }
    RasterDataWidthInPixels->num = header.cupsWidth;

    RasterDataHeightInPixels = get_attribute("RasterDataHeightInPixels");
    if (NULL == RasterDataHeightInPixels)
    {
        RasterDataHeightInPixels = (struct tagAttribute*)malloc(sizeof(*RasterDataHeightInPixels));
        RasterDataHeightInPixels->node.type = 'a';
        RasterDataHeightInPixels->type = GPD_NUMBER;
        RasterDataHeightInPixels->name = (char*)"RasterDataHeightInPixels";
        cache_attribute(RasterDataHeightInPixels);
    }
    RasterDataHeightInPixels->num = header.cupsHeight;

    PageNumber = get_attribute("PageNumber");
    if (NULL == PageNumber)
    {
        PageNumber = (struct tagAttribute*)malloc(sizeof(*PageNumber));
        PageNumber->node.type = 'a';
        PageNumber->type = GPD_NUMBER;
        PageNumber->name = (char*)"PageNumber";
        cache_attribute(PageNumber);
    }

    _cupsLangPrintFilter(stderr, "DEBUG", "Start page loop");

    /*
    * Process pages as needed...
    */
    page = 0;
    margin_x = (int)(current_dpi.first/72.0f  * header.Margins[0]);
    margin_y = (int)(current_dpi.second/72.0f * header.Margins[1]);

    do
    {
        if (g_Canceled)
            break;

        PageNumber->num = ++page;

        /*
        * Write a status message with the page number and number of copies.
        */
        _cupsLangPrintFilter(stderr, "INFO", _("Starting page %d, %d"),
                page, header.NumCopies);

        switch(language)
        {
            case PL_PCL_XL:
            ret = convert_raster_block_pclxl(ras,
                        &header,
                        &raster_block,
                        &block_count);
            break;

            case PL_ESC_P:
            case PL_ESC_P_2:
            ret = convert_raster_block(ras,
                    &header,
                    Black_1_BPP,
                    h_byte,
                    interlacing,
                    block_lines,
                    bud_header.sub_blocks,
                    &raster_block,
                    &block_count);
                    break;
            default:
                break;
        }
        if (ret < 0)
        {
            _cupsLangPrintFilter(stderr, "ERROR", _("raster convert failed %d"), ret);
            break;
        }

        StartPage(ppd, &header);

        last_y = 0;
        last_x = 0;

        for (int i = 0; i<block_count; ++i)
        {
            if (!raster_block[i].data_bytes) continue;

            /*convert pixel unit to printer's master unit*/
            int destx = (raster_block[i].x + margin_x) * master_unit.first / current_dpi.first;
            int desty = (raster_block[i].y + margin_y) * master_unit.second / current_dpi.second;

            if (CmdCR)
            {
                execute_cmd_expr(CmdCR, NULL);
            }

            if (CmdYMoveAbsolute)
            {
                y_attr->num = desty;
                execute_cmd_expr(CmdYMoveAbsolute, NULL);
            }
            else if (CmdYMoveRelDown)
            {
                y_attr->num = desty - last_y;
                execute_cmd_expr(CmdYMoveRelDown, NULL);
            }
            last_y = desty;

            if (destx > 0)
            {
                if (CmdXMoveAbsolute)
                {
                    x_attr->num = destx;
                    execute_cmd_expr(CmdXMoveAbsolute, NULL);
                }
                else if (CmdXMoveRelRight && CmdXMoveRelLeft)
                {
                    if (destx > last_x)
                    {
                        x_attr->num = destx - last_x;
                        execute_cmd_expr(CmdXMoveRelRight, NULL);
                    }
                    else
                    {
                        x_attr->num = last_x - destx;
                        execute_cmd_expr(CmdXMoveRelLeft, NULL);
                    }
                }
            }
            last_x = destx;

            if (CmdBeginRaster) execute_cmd_expr(CmdBeginRaster, NULL);

            NumOfDataBytes->num = raster_block[i].data_bytes;
            execute_cmd_expr(CmdSendBlockData, &raster_block[i]);

            write_printer(raster_block[i].data, raster_block[i].data_bytes);

            if (CmdEndRaster) execute_cmd_expr(CmdEndRaster, NULL);
        }

        _cupsLangPrintFilter(stderr, "INFO", _("Finished page %d."), page);
        EndPage(&header);

        if (g_Canceled)
          break;
    } while (cupsRasterReadHeader2(ras, &header));

    Shutdown();
    ppdClose(ppd);
    cupsRasterClose(ras);
    if (fd != 0)
    {
       close(fd);
    }

    uninit_log();
    if (g_var_map) delete g_var_map;

    /*
    * If no pages were printed, send an error message...
    */
    if (page == 0)
    {
        _cupsLangPrintFilter(stderr, "ERROR", _("No pages were found."));
        return (1);
    }

    return (0);
}

/*end*/
